#   version:    1.0     First pass at PyConTurb
#               2.0     Switch to uvw, other changes
#                2.5    Fixed bug in custom profiles
#               2.6     Condata is deprecated.
#               2.7     Write coherence to file
__version__ = '2.7.1'